#ifndef __SYNC_H__
#define __SYNC_H__

#include <assert.h>

#ifdef _WIN32

#include <windows.h>

#define thread_proc WINAPI

class thread_t {
    HANDLE h;
  public:
    typedef void (thread_proc* thread_proc_t)(void*);

    void start(thread_proc_t f, void* arg) {
	DWORD threadid;
	h = CreateThread(NULL, NULL, LPTHREAD_START_ROUTINE(f), arg,
			 0, &threadid);
    }
    void join() {
	WaitForSingleObject(h, INFINITE);
	CloseHandle(h);
	h = NULL;
    }
};

class semaphore_t {
    HANDLE s;
  public:
    void wait() {
	int rc = WaitForSingleObject(s, INFINITE);
	assert(rc == WAIT_OBJECT_0 || rc == WAIT_TIMEOUT);
    }
    void signal(unsigned inc = 1) {
	if (inc != 0) {
	    ReleaseSemaphore(s, inc, NULL);
	}
    }
    semaphore_t(unsigned initValue = 0) {
	s = CreateSemaphore(NULL, initValue, 1000000, NULL);
	assert(s != NULL);
    }
    ~semaphore_t() {
	CloseHandle(s);
    }
};

#else

#define thread_proc

#include <pthread.h>

class thread_t { 
    pthread_t thread;
  public:
    typedef void (thread_proc* thread_proc_t)(void*);

    void start(thread_proc_t f, void* arg) {
#if defined(_AIX41)
	pthread_attr_t attr;
	pthread_attr_init(&attr);
	// At AIX 4.1, 4.2 threads are by default created detached
	pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_UNDETACHED);
	pthread_create(&thread, &attr, (void*(*)(void*))f, arg);
	pthread_attr_destroy(&attr);
#else
	pthread_create(&thread, NULL, (void*(*)(void*))f, arg);
#endif
    }

    void join() {
	void* result;
	pthread_join(thread, &result);
    }
};

class semaphore_t {
    pthread_cond_t   cond;
    pthread_mutex_t  mutex;
    int              count;
  public:
    void wait() {
	pthread_mutex_lock(&mutex);
	while (count == 0) {
	    pthread_cond_wait(&cond, &mutex);
	}
	count -= 1;
	pthread_mutex_unlock(&mutex);
    }

    void signal(unsigned inc = 1) {
	pthread_mutex_lock(&mutex);
	count += inc;
	if (inc > 1) {
	    pthread_cond_broadcast(&cond);
	} else if (inc == 1) {
	    pthread_cond_signal(&cond);
	}
	pthread_mutex_unlock(&mutex);
    }

    semaphore_t(unsigned initValue = 0) {
	pthread_cond_init(&cond, NULL);
	pthread_mutex_init(&mutex, NULL);
	count = initValue;
    }
    ~semaphore_t() {
	pthread_mutex_destroy(&mutex);
	pthread_cond_destroy(&cond);
    }
};

#endif


#endif
